#include "../../internal-include/CD/cd_internal.hpp"
#include <PSX/System/IOPorts/dma_io.hpp>
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.hpp>
#include <stdio.hpp>

namespace JabyEngine {
    namespace CD {
        namespace internal {
            union Configuration {
                struct File {
                    uint32_t cur_lba;
                    uint32_t dst_lba;

                    void set_from(const AutoLBAEntry& file_info) {
                        this->cur_lba = file_info.get_lba();
                        this->dst_lba = this->cur_lba + file_info.get_size_in_sectors();
                    }

                    bool done_processing() {
                        this->cur_lba++;
                        return this->cur_lba == this->dst_lba;
                    }
                };

                struct XA {
                    CDTimeStamp start_time;
                    uint8_t     channel;
                };

                File file;
                XA   xa;
            };

            static constexpr auto AudioSectorMode   = CD_IO::Mode::from(CD_IO::Mode::SingleSpeed, CD_IO::Mode::AutoPauseTrack, CD_IO::Mode::CDDA);
            static constexpr auto DataSectorMode    = CD_IO::Mode::from(CD_IO::Mode::DoubleSpeed, CD_IO::Mode::DataSector);
            static constexpr auto XAAudioSectorMode = CD_IO::Mode::from(CD_IO::Mode::SingleSpeed, CD_IO::Mode::XADPCM, CD_IO::Mode::WholeSector, CD_IO::Mode::UseXAFilter);
            
            namespace IRQ {
                static SysCall::InterruptVerifierResult verifier();
                static uint32_t                         handler(uint32_t);
            }

            static SectorBufferAllocator sector_allocator;
            static Configuration         cur_cfg;
            
            uint8_t                   cmd_interrupt_bit = 0;
            State                     current_state     = State::Ready;
            SysCall::InterrupCallback callback          = {
                .next              = nullptr,
                .handler_function  = IRQ::handler,
                .verifier_function = IRQ::verifier
            };

            // Requires Index0
            static CDTimeStamp send_read_cmd(uint32_t lba, CD_IO::Command::Desc desc) {
                const auto loc = CDTimeStamp::from(lba);               

                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, loc.get_min_cd(), loc.get_sec_cd(), loc.get_sector_cd());
                Command::send<CD_IO::PortIndex0>(desc);
                return loc;
            }

            static void send_read_n(uint32_t lba) {
                send_read_cmd(lba, CD_IO::Command::ReadN);
                current_state = State::Reading;
            }

            namespace IRQ {
                static void read_sector_dma(uint32_t* dst, size_t bytes) {
                    static const auto WaitSectorReady = []() {
                        while(!CD_IO::IndexStatus.read().is_set(CD_IO::IndexStatus::HasDataFifoData));
                    };

                    static const auto ReadSector = [](uint32_t* dst, size_t bytes) {
                        DMA_IO::CDROM.set_adr(reinterpret_cast<uintptr_t>(dst));
                        DMA_IO::CDROM.block_ctrl.write(DMA_IO::BCR::SyncMode0::for_cd(bytes >> 2));
                        DMA_IO::CDROM.channel_ctrl.write(DMA_IO::CHCHR::StartCDROM());
                        
                        DMA_IO::CDROM.wait();

                        CD_IO::PortIndex0::Request.write(CD_IO::Request::reset());
                    };

                    WaitSectorReady();
                    ReadSector(dst, bytes);
                }

                static void read_sector_to(uint32_t* dst, size_t bytes) {
                    CD_IO::PortIndex0::change_to();
                    CD_IO::PortIndex0::Request.write(CD_IO::Request::want_data());

                    // We only support DMA rn
                    read_sector_dma(dst, bytes);

                    // Do we ever want to support reading via IO Port?
                    // Doesn't seem to important when we can use DMA
                }

                static void resume_at(const CDTimeStamp& cd_time) {
                    CD_IO::PortIndex0::change_to();
                    Command::send<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, cd_time.get_min_cd(), cd_time.get_sec_cd(), cd_time.get_sector_cd());

                    CD_IO::PortIndex1::change_to();
                    while(CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag) != CD_IO::Interrupt::Acknowledge);
                    CD_IO::Interrupt::ack_extended(CD_IO::PortIndex1::InterruptFlag);

                    CD_IO::PortIndex0::change_to();
                }

                //######################################################################################################################

                static SysCall::InterruptVerifierResult verifier() {
                    if(Interrupt::is_irq(Interrupt::CDROM)) {
                        return SysCall::InterruptVerifierResult::ExecuteHandler;
                    }
                    
                    else {
                        return SysCall::InterruptVerifierResult::SkipHandler;
                    }
                }

                static uint32_t handler(uint32_t) {
                    const auto old_status = CD_IO::IndexStatus.read();

                    CD_IO::PortIndex1::change_to();
                    const auto cur_irq = CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag);
                    CD_IO::Interrupt::ack_extended(CD_IO::PortIndex1::InterruptFlag);

                    cmd_interrupt_bit = bit::clear(cmd_interrupt_bit, cur_irq);

                    if(current_state != State::XAMode) {
                        switch(cur_irq) {
                            case CD_IO::Interrupt::DataReady: {
                                // Obtain sector content here
                                auto* sector = sector_allocator.allocate_sector();
                                if(sector) {
                                    //Now obtain sector
                                    read_sector_to(sector->data, CD_IO::DataSector::SizeBytes);

                                    if(cur_cfg.file.done_processing()) {
                                        current_state = State::Done;
                                        pause();
                                    }
                                }

                                else {
                                    current_state = State::BufferFull;
                                    pause();
                                }
                            } break;

                            case CD_IO::Interrupt::DataEnd: {
                                // TODO: Fix this!! This is a freaking static time
                                resume_at(CDTimeStamp{.min = 0, .sec = 9, .sector = 0});
                                Command::send<CD_IO::PortIndex0>(CD_IO::Command::Play);
                            } break;

                            case CD_IO::Interrupt::DiskError: {
                                current_state = State::Error;
                            } break;
                        }
                    }

                    else {
                        switch(cur_irq) {
                            case CD_IO::Interrupt::DataReady: {
                                XASectorHeader xa_file;

                                read_sector_to(reinterpret_cast<uint32_t*>(&xa_file), sizeof(XASectorHeader));
                                if(cur_cfg.xa.channel == xa_file.sub_header.channel_number) {
                                    resume_at(cur_cfg.xa.start_time);
                                    Command::send<CD_IO::PortIndex0>(CD_IO::Command::ReadS);
                                }
                            } break;

                            case CD_IO::Interrupt::DiskError: {
                                current_state = State::Error;
                            } break;
                        };
                    }

                    // No masking required because we can only write bit 0 - 2
                    CD_IO::IndexStatus.write(old_status);
                    Interrupt::ack_irq(Interrupt::CDROM);
                    SysCall::ReturnFromException();
                    __builtin_unreachable();
                }
            }

            void read_file(AutoLBAEntry file_info, const SectorBufferAllocator& buffer_allocator) {
                cur_cfg.file.set_from(file_info);
                sector_allocator = buffer_allocator;

                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, DataSectorMode);

                send_read_n(cur_cfg.file.cur_lba);
            }

            void read_xa(uint32_t lba, uint8_t channel) {
                static constexpr uint8_t File = 1;

                cur_cfg.xa.channel   = channel;
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::Filter, File, channel);

                cur_cfg.xa.start_time = send_read_cmd(lba, CD_IO::Command::ReadS);
                current_state = State::XAMode;
            }

            void continue_reading() {
                if(current_state == State::BufferFull) {
                    Command::wait_completed();
                    send_read_n(cur_cfg.file.cur_lba);    
                }                
            }

            CDTimeStamp get_lock() {
                Command::wait_completed();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::GetLocP);

                const auto track  = CD_IO::PortIndex0::ResponseFifo.read().raw; // track number (AAh=Lead-out area) (FFh=unknown, toc, none?)
                const auto index  = CD_IO::PortIndex0::ResponseFifo.read().raw; // index number (Usually 01h)
                const auto mm     = CD_IO::PortIndex0::ResponseFifo.read().raw; // minute number within track (00h and up)
                const auto ss     = CD_IO::PortIndex0::ResponseFifo.read().raw; // second number within track (00h to 59h)
                const auto sect   = CD_IO::PortIndex0::ResponseFifo.read().raw; // sector number within track (00h to 74h)

                return CDTimeStamp{
                    .min    = CD_IO::PortIndex0::ResponseFifo.read().raw, // minute number on entire disk (00h and up)
                    .sec    = CD_IO::PortIndex0::ResponseFifo.read().raw, // second number on entire disk (00h to 59h)
                    .sector = CD_IO::PortIndex0::ResponseFifo.read().raw, // sector number on entire disk (00h to 74h)                    
                };
            }

            void enable_CDDA() {
                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, AudioSectorMode);
            }

            void enable_CDXA(bool double_speed) {
                static constexpr uint8_t SingleSpeedBit = 0x0;
                static constexpr uint8_t DoubleSpeedBit = static_cast<uint8_t>(CD_IO::Mode::DoubleSpeed);

                const uint8_t mode = XAAudioSectorMode.raw | (double_speed ? DoubleSpeedBit : SingleSpeedBit);

                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, mode);
            }
        }
    }
}